import os

import pytest


TEST_CONFIG_DIR = os.path.join(
    os.path.dirname(os.path.realpath(__file__)), "./qcs_config"
)
TEST_QCS_SETTINGS_PATH = os.path.join(TEST_CONFIG_DIR, "settings.toml")
TEST_QCS_SECRETS_PATH = os.path.join(TEST_CONFIG_DIR, "secrets.toml")
os.environ["QCS_SETTINGS_FILE_PATH"] = TEST_QCS_SETTINGS_PATH
os.environ["QCS_SECRETS_FILE_PATH"] = TEST_QCS_SECRETS_PATH


@pytest.fixture
def device_2q() -> str:
    import json

    return json.dumps(
        {
            "isa": {
                "1Q": {
                    "0": {
                        "id": 0,
                        "gates": [
                            {
                                "operator": "RX",
                                "duration": 50.0,
                                "fidelity": 1.0,
                                "parameters": [0.0],
                                "arguments": [0],
                                "operator_type": "gate",
                            },
                            {
                                "operator": "RX",
                                "duration": 50.0,
                                "fidelity": 0.9909074679565163,
                                "parameters": [3.141592653589793],
                                "arguments": [0],
                                "operator_type": "gate",
                            },
                            {
                                "operator": "RX",
                                "duration": 50.0,
                                "fidelity": 0.9909074679565163,
                                "parameters": [-3.141592653589793],
                                "arguments": [0],
                                "operator_type": "gate",
                            },
                            {
                                "operator": "RX",
                                "duration": 50.0,
                                "fidelity": 0.9909074679565163,
                                "parameters": [1.5707963267948966],
                                "arguments": [0],
                                "operator_type": "gate",
                            },
                            {
                                "operator": "RX",
                                "duration": 50.0,
                                "fidelity": 0.9909074679565163,
                                "parameters": [-1.5707963267948966],
                                "arguments": [0],
                                "operator_type": "gate",
                            },
                            {
                                "operator": "RZ",
                                "duration": 0.01,
                                "fidelity": 1.0,
                                "parameters": ["_"],
                                "arguments": [0],
                                "operator_type": "gate",
                            },
                            {
                                "operator": "MEASURE",
                                "duration": 2000.0,
                                "fidelity": 0.977,
                                "qubit": 0,
                                "target": "_",
                                "operator_type": "measure",
                            },
                            {
                                "operator": "MEASURE",
                                "duration": 2000.0,
                                "fidelity": 0.977,
                                "qubit": 0,
                                "target": None,
                                "operator_type": "measure",
                            },
                        ],
                    },
                    "1": {
                        "id": 1,
                        "gates": [
                            {
                                "operator": "RX",
                                "duration": 50.0,
                                "fidelity": 1.0,
                                "parameters": [0.0],
                                "arguments": [1],
                                "operator_type": "gate",
                            },
                            {
                                "operator": "RX",
                                "duration": 50.0,
                                "fidelity": 0.9967034552975036,
                                "parameters": [3.141592653589793],
                                "arguments": [1],
                                "operator_type": "gate",
                            },
                            {
                                "operator": "RX",
                                "duration": 50.0,
                                "fidelity": 0.9967034552975036,
                                "parameters": [-3.141592653589793],
                                "arguments": [1],
                                "operator_type": "gate",
                            },
                            {
                                "operator": "RX",
                                "duration": 50.0,
                                "fidelity": 0.9967034552975036,
                                "parameters": [1.5707963267948966],
                                "arguments": [1],
                                "operator_type": "gate",
                            },
                            {
                                "operator": "RX",
                                "duration": 50.0,
                                "fidelity": 0.9967034552975036,
                                "parameters": [-1.5707963267948966],
                                "arguments": [1],
                                "operator_type": "gate",
                            },
                            {
                                "operator": "RZ",
                                "duration": 0.01,
                                "fidelity": 1.0,
                                "parameters": ["_"],
                                "arguments": [1],
                                "operator_type": "gate",
                            },
                            {
                                "operator": "MEASURE",
                                "duration": 2000.0,
                                "fidelity": 0.9450000000000001,
                                "qubit": 1,
                                "target": "_",
                                "operator_type": "measure",
                            },
                            {
                                "operator": "MEASURE",
                                "duration": 2000.0,
                                "fidelity": 0.9450000000000001,
                                "qubit": 1,
                                "target": None,
                                "operator_type": "measure",
                            },
                        ],
                    },
                },
                "2Q": {
                    "0-1": {
                        "ids": [0, 1],
                        "gates": [
                            {
                                "operator": "CZ",
                                "duration": 200.0,
                                "fidelity": 0.95,
                                "parameters": [],
                                "arguments": ["_", "_"],
                                "operator_type": "gate",
                            },
                        ],
                    },
                },
            },
            "specs": {},
        }
    )


@pytest.fixture
def native_bitflip_program() -> str:
    return """
DECLARE ro BIT[0]
RX(pi) 0
MEASURE 0 ro[0]
"""


@pytest.fixture
def bell_program() -> str:
    return """
DECLARE ro BIT[2]
X 0
CNOT 0 1
MEASURE 0 ro[0]
MEASURE 1 ro[1]
"""



"""
The qcs_sdk module provides an interface to Rigetti Quantum Cloud Services. Allowing users to compile and run Quil programs on Rigetti quantum processors.
"""
from typing import Any, Dict, List, Optional, TypedDict
from numbers import Number

RecalculationTable = List[str]
Memory = Dict[str, List[float]]
PatchValues = Dict[str, List[float]]

class RewriteArithmeticResults(TypedDict):
    program: str
    """
    The resulting program where gate parameter arithmetic has been replaced with memory references. Before execution, the program memory should be updated using the `recalculation_table`.
    """

    recalculation_table: RecalculationTable
    """ 
    The recalculation table stores an ordered list of arithmetic expressions, which are to be used when updating the program memory before execution.
    """

class TranslationResult(TypedDict):
    memory_descriptors: Optional[Dict[str, Any]]
    """
    A map from the name of memory (declared with `DECLARE`) to the size and type of that memory.
    """

    program: str
    """
    The compiled program binary.
    """

    ro_sources: Optional[List[List[str]]]
    """
    A mapping from the program's memory references to the key used to index the results map.
    """

    settings_timestamp: Optional[str]
    """
    The timestamp of the settings used during translation.
    """

class ExecutionResult(TypedDict):
    shape: List[int]
    """
    The shape of the result data.
    """

    data: List[Number | List[float]]
    """
    The result data. Complex numbers are represented as [real, imaginary].
    """

    dtype: str
    """
    The type of the result data (as a `numpy` `dtype`).
    """

class ExecutionResults(TypedDict):
    buffers: Dict[str, ExecutionResult]
    """
    The readout results of execution, mapping a published filter node to its data.

    See `TranslationResult.ro_sources` which provides the mapping from the filter node name to the name of the memory declaration in the source program.
    """

    execution_duration_microseconds: Optional[int]
    """
    The time spent executing the program.
    """

async def compile(quil: str, target_device: str, *, timeout: int = 30) -> str:
    """
    Uses quilc to convert a quil program to native Quil.

    Args:
        quil: A Quil program.
        target_device: A JSON encoded description of the Quantum Abstract Machine Architecture.

    Keyword Args:
        timeout: The number of seconds to wait before timing out. If set to None, there is no timeout (default: 30).

    Returns:
        An Awaitable that resolves to the native Quil program.
    """
    ...

def rewrite_arithmetic(native_quil: str) -> RewriteArithmeticResults:
    """
    Rewrites parametric arithmetic such that all gate parameters are only memory references
    to a newly declared memory location (__SUBST).

    Args:
        native_quil: A Quil program.

    Returns:
        A dictionary with the rewritten program and recalculation table (see `RewriteArithmeticResults`).
    """
    ...

def build_patch_values(
    recalculation_table: RecalculationTable, memory: Memory
) -> PatchValues:
    """
    Evaluate the expressions in recalculation_table using the numeric values
    provided in memory.

    Args:
        recalculation_table: an ordered list of arithmetic expressions, which are to be used when updating the program memory before execution (see `rewrite_arithmetic`).
        memory: A mapping of symbols to their values (e.g. `{"theta": [0.0]}`).

    Returns:
        A dictionary that maps each symbol to the value it should be patched with.
    """
    ...

async def translate(
    native_quil: str, num_shots: int, quantum_processor_id: str
) -> TranslationResult:
    """
    Translates a native Quil program into an executable.

    Args:
        native_quil: A Quil program.
        num_shots: The number of shots to perform.
        quantum_processor_id: The ID of the quantum processor the executable will run on (e.g. "Aspen-M-2").

    Returns:
        An Awaitable that resolves to a dictionary with the compiled program, memory descriptors, and readout sources (see `TranslationResult`).
    """
    ...

async def submit(
    program: str, patch_values: Dict[str, List[float]], quantum_processor_id: str
) -> str:
    """
    Submits an executable `program` to be run on the specified QPU.

    Args:
        program: An executable program (see `translate`).
        patch_values: A mapping of symbols to their desired values (see `build_patch_values`).
        quantum_processor_id: The ID of the quantum processor to run the executable on.

    Returns:
        An Awaitable that resolves to the ID of the submitted job.
    """
    ...

async def retrieve_results(job_id: str, quantum_processor_id: str) -> ExecutionResults:
    """
    Fetches results for the corresponding job ID.

    Args:
        job_id: The ID of the job to retrieve results for.
        quantum_processor_id: The ID of the quanutum processor the job ran on.

    Returns:
        An Awaitable that resolves to a dictionary describing the results of the execution and its duration (see `ExecutionResults`).
    """
    ...

async def get_quilc_version() -> str:
    """
    Returns the version number of the running quilc server.
    """
    ...



from unittest.mock import patch
import pytest

import re

import qcs_sdk


@pytest.mark.asyncio
async def test_compile(native_bitflip_program: str, device_2q: str):
    await qcs_sdk.compile(native_bitflip_program, device_2q)


@pytest.mark.asyncio
@pytest.mark.skip
async def test_translate(native_bitflip_program: str, device_2q: str):
    await qcs_sdk.translate(native_bitflip_program, 1, "Aspen-11")


@pytest.mark.asyncio
@pytest.mark.skip
async def test_execute(bell_program: str, device_2q):
    compiled_program = await qcs_sdk.compile(bell_program, device_2q)
    translation_result = await qcs_sdk.translate(compiled_program, 1, "Aspen-11")
    job_id = await qcs_sdk.submit(translation_result["program"], {}, "Aspen-11")
    await qcs_sdk.retrieve_results(job_id, "Aspen-11")


def test_rewrite_arithmetic():
    native_quil = "RX(2*theta[0]) 0"
    rewritten_arithmetic = qcs_sdk.rewrite_arithmetic(native_quil)
    expected = {
        "program": "DECLARE __SUBST REAL[1]\nRX(__SUBST[0]) 0\n",
        "recalculation_table": ["((2*theta[0])/6.283185307179586)"],
    }
    assert rewritten_arithmetic == expected


def test_build_patch_values():
    memory = {"theta": [2.0]}
    recalculation_table = ["0.5*theta[0]"]
    expected = {"theta": [2.0], "__SUBST": [1.0]}
    patch_values = qcs_sdk.build_patch_values(recalculation_table, memory)
    assert patch_values == expected


@pytest.mark.asyncio
async def test_get_quilc_version():
    version = await qcs_sdk.get_quilc_version()
    assert re.match(r"^([0-9]+)\.([0-9]+)\.([0-9]+)$", version)
